#include <XBee.h>
#include <SoftwareSerial.h>
XBeeWithCallbacks xbee;

// Software serial pins
#define xb_rx 7
#define xb_tx 6
#define xb_rst 8

SoftwareSerial nss(xb_rx,xb_tx);
AtCommandRequest atCommands[10];

void setup() {
  Serial.begin(9600);
  
  log("Program start");
  log("reset xbee");
  //reset xbee
  pinMode(xb_rst,LOW);
  digitalWrite(xb_rst, LOW);
  digitalWrite(xb_rst, HIGH);
    
  
  xbee.setSerial(nss);
  xbee.onAtCommandResponse(atCmdResp);
  
  //Set API mode 2
  sendAtCmd('AP',2);
  
  //Set scan channels 
  // https://www.digi.com/resources/documentation/Digidocs/90002002/Default.htm#Reference/r_cmd_SC.htm?TocPath=AT%2520commands%257CNetwork%2520commands%257C_____2
  sendAtCmd('SC',1);

  //Enable encryption
  // https://www.digi.com/resources/documentation/Digidocs/90002002/Default.htm#Reference/r_cmd_EE.htm?TocPath=AT%2520commands%257CSecurity%2520commands%257C_____1
  sendAtCmd('EE',1);

  //Set link key
  // https://www.digi.com/resources/documentation/Digidocs/90002002/Default.htm#Reference/r_cmd_KY.htm?TocPath=AT%2520commands%257CSecurity%2520commands%257C_____3
  sendAtCmd('KY',NULL);

  //Write changes
  sendAtCmd('WR',NULL);
}

void loop() {
  // put your main code here, to run repeatedly:
  log("AT setup done");
  delay(10000);
}

void atCmdResp(AtCommandResponse& resp, uintptr_t){
  log("Got AT Command response");
  if (resp.getStatus() == AT_OK){
  }
}

bool sendAtCmd(char* cmd, uint8_t* parameter){
  AtCommandRequest atRequest = AtCommandRequest();
  atRequest.setCommand(cmd);
  if (parameter != NULL) atRequest.setCommandValue(parameter);
  log("Sending AT command");
  xbee.send(atRequest);
  bool success = false;
  bool wait = true;
  if (wait){
    // Wait for a response from the sent command
    AtCommandResponse atResponse = AtCommandResponse();
    while (success == 0){
      if (xbee.readPacket(5000)){
        if (xbee.getResponse().getApiId() == AT_COMMAND_RESPONSE) {
          xbee.getResponse().getAtCommandResponse(atResponse);
          if (atResponse.getCommand()[0] == cmd[0] &&
              atResponse.getCommand()[1] == cmd[1]) {
            //got the correct command response
            if (atResponse.isOk()){
              success = 1;
              String str;
              for (int i=0;i<atResponse.getValueLength();i++){
                str = str + (atResponse.getValue()[i],HEX);  
              }
              log(str);
            }else{
              //Response to right command, but AT command failed
            }
          }else{
            // Response to wrong command
            log("Got response, but wrong command: "+ (atResponse.getCommand()[0],HEX)+ (atResponse.getCommand()[1],HEX));
          }
        }else{
          //Not an AT command response
          log("Expexted AT, got "+(xbee.getResponse().getApiId(), HEX)); 
        }
        
      }else{
        log("No packet from XBee. Retrying...");
      }
    }
  }
  return true;
}

void log(String str){
  Serial.println(millis()+": "+str);
}
